require 'eventmachine'
require 'escape_utils'
require 'addressable/uri'

require 'base64'
require 'socket'

require 'http11_client'
require 'em_buffer'

require 'em-http/core_ext/bytesize'
require 'em-http/http_header'
require 'em-http/http_encoding'
require 'em-http/http_options'
require 'em-http/client'
require 'em-http/multi'
require 'em-http/request'
require 'em-http/decoders'
require 'em-http/mock'